/*
** -----------------------------------------------------------------------------**
** ccd_seq.v
**
** Sequencer combined with program memory (downloadable in run time)
**
** Copyright (C) 2003-2006 Elphel, Inc.
**
** -----------------------------------------------------------------------------**
**  This file is part of x324
**  x324 is free software - hardware description language (HDL) code.
** 
**  This program is free software: you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation, either version 3 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program.  If not, see <http://www.gnu.org/licenses/>.
** -----------------------------------------------------------------------------**
**
**  $Log: ccd_seq.v,v $
**  Revision 1.1.1.1  2008/11/27 20:04:02  elphel
**
**
**  Revision 1.1.1.1  2007/08/17 12:07:19  elphel
**  This is a fresh tree based on elphel353-2.10
**
**  Revision 1.6  2007/08/17 12:07:19  spectr_rain
**  switch to GPL3 license
**
**  Revision 1.5  2007/07/20 09:46:13  spectr_rain
**  *** empty log message ***
**
**  Revision 1.5  2007/05/25 02:09:09  elphel
**  347 update for mechanical shutter
**
**  Revision 1.3  2007/04/06 16:45:33  elphel
**  modifying mechanical shutter control
**
**  Revision 1.1  2006/09/24 16:44:36  elphel
**  Split file into separate module files, added real time clock module, project files
**
*/
//
// this module runs at pixel clock (25MHz?), output signals will be generated/synchronized with fast clock (125MHz)
// for now only fast, single-frame readout with hard-wired frame size, no binning.
// exposure up to 6.5535 sec - 16 bits, 100usec step

module ccd_seq (clk, // pixel clock
                wr,  // write enable, sync to clk
                wa,  // wa
                wd,  // 16 bit data (now - only exposure time)
                en,  // enable sequencer
                strt, // start exposure, when finished - readout
                en_h,   // enable horizontal phases
                v_ph, // [2:0] - number of prewritten vertical phases combination
                v_stb,   // change vertical phases to v_phase,
                fd,      // activate "fast drain"
                shtr,    // activate shutter
                frame, // frame active
                line,    // en_h masked (by frame) all latency will be added externally
                expos,   // sequencer status (0- slow erase, 1 - fast erase, 2 - exposure, 3 - readout.
                shtrmode, // 1 - use mechanical shutter (first pair of toggles)
                mshutter  // mechanical shutter pulse output
                );
//  parameter NLINES=2672;
//  parameter NPIX=  2040; // total pixels in a half-line (with black ones)
//  parameter T_HD=   100;   //  4.0 us (3.0/3.5/10.0), HCCD delay
//  parameter T_VCCD= 100;   //  4.0 us (3.0/3.5/20.0), VCCD transfer time
//  parameter T_V3RD= 250;   // 10.0 us (8.0/10.0/15.0), Photodiode transfer time
//  parameter T_V3P= 1500;   // 60.0 us (50.0/60.0/80.0), Photodiode transfer time
//  parameter T_V3D=  500;   // 20.0 us (15.0/20.0/80.0), VCCD delay
//  parameter T_S=    125;   //  5.0 us (3.0/4.0/10.0), Shutter pulse time
//  parameter T_SD=    50;   //  2.0 us (1.0/1.5/10.0), Shutter delay time
//  parameter T_FD=    25;   //  1.0 us (0.5/?/?) Fast Dump Gate delay


  input         clk;
  input         wr;
  input   [7:0] wa;
  input  [15:0] wd;
  input         en;
  input         strt;
  output        en_h;
  output  [2:0] v_ph;
  output        v_stb;
  output        fd;
  output        shtr;
  output        frame;
  output        line;
  output        expos;
  input  shtrmode;
  output mshutter;


  wire   [7:0] pgm_a;
  wire  [15:0] pgm_d;
  wire         pgm_re;
  wire         line=frame && en_h;
/*
  RAMB4_S16_S16 i_seqpgm (.DOA(),
                          .ADDRA(wa[7:0]),
                          .CLKA(clk),
                          .DIA(wd[15:0]),
                          .ENA(wr),
                          .RSTA(1'b0),
                          .WEA(1'b1),
                          .DOB(pgm_d[15:0]),
                          .ADDRB(pgm_a[7:0]),
                          .CLKB(clk),
                          .DIB(16'b0),
                          .ENB(pgm_re),
                          .RSTB(1'b0),
                          .WEB(1'b0));
*/
 RAMB16_S18_S18 i_seqpgm (
      .DOA(),      // Port A 16-bit Data Output
      .DOPA(),    // Port A 2-bit Parity Output
      .ADDRA({2'h0,wa[7:0]}),  // Port A 10-bit Address Input
      .CLKA(clk),    // Port A Clock
      .DIA(wd[15:0]),      // Port A 16-bit Data Input
      .DIPA(2'h0),    // Port A 2-bit parity Input
      .ENA(wr),      // Port A RAM Enable Input
      .SSRA(1'b0),    // Port A Synchronous Set/Reset Input
      .WEA(1'b1),      // Port A Write Enable Input

      .DOB(pgm_d[15:0]),      // Port B 16-bit Data Output
      .DOPB(),    // Port B 2-bit Parity Output
      .ADDRB({2'h0,pgm_a[7:0]}),  // Port B 10-bit Address Input
      .CLKB(clk),    // Port B Clock
      .DIB(16'h0),      // Port B 16-bit Data Input
      .DIPB(2'h0),    // Port-B 2-bit parity Input
      .ENB(pgm_re),      // Port B RAM Enable Input
      .SSRB(1'b0),    // Port B Synchronous Set/Reset Input
      .WEB(1'b0)       // Port B Write Enable Input
   );

  seq_cpu  i_seq_cpu (.clk(clk),  // clock
                      .en( en),  // sync reset to address 0, stays there if held
                      .strt(strt), // exposure start, actual may be delayed
                      .a(pgm_a[7:0]),    // [ 7:0] external program memory address
                      .re(pgm_re),   // enable program memory read (0 - hold)
                      .d(pgm_d[15:0]),    // [15:0] external program memory data
                      .v_ph(v_ph), // [ 2:0] vertical phase
                      .v_stb(v_stb),// vertical phase changed
                      .expos(expos),// exposure in progress
                      .frame(frame),// frame readout in progress
//                    .line(line), // line readout in progress (active hor. phases)
                      .line(en_h), // line readout in progress (active hor. phases)
                      .fd(fd),   // fast drain
                      .shtr(shtr),// shutter
                      .shtrmode(shtrmode), // 1 - use mechanical shutter (first pair of toggles)
                      .mshutter(mshutter)  // mechanical shutter pulse output
                      );
endmodule
