/*
** -----------------------------------------------------------------------------**
** seq_cpu.v
**
** Programmable sequencer for the CCD control signals
**
** Copyright (C) 2003-2006 Elphel, Inc.
**
** -----------------------------------------------------------------------------**
**  This file is part of x324
**  x324 is free software - hardware description language (HDL) code.
** 
**  This program is free software: you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation, either version 3 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program.  If not, see <http://www.gnu.org/licenses/>.
** -----------------------------------------------------------------------------**
**
**  $Log: seq_cpu.v,v $
**  Revision 1.1.1.1  2008/11/27 20:04:02  elphel
**
**
**  Revision 1.1.1.1  2007/08/17 12:07:19  elphel
**  This is a fresh tree based on elphel353-2.10
**
**  Revision 1.6  2007/08/17 12:07:19  spectr_rain
**  switch to GPL3 license
**
**  Revision 1.5  2007/07/20 09:46:13  spectr_rain
**  *** empty log message ***
**
**  Revision 1.4  2007/05/25 02:09:09  elphel
**  347 update for mechanical shutter
**
**  Revision 1.3  2007/04/06 16:45:33  elphel
**  modifying mechanical shutter control
**
**  Revision 1.1  2006/09/24 16:44:36  elphel
**  Split file into separate module files, added real time clock module, project files
**
*/
module seq_cpu     (clk, // clock
                    en,  // sync reset to address 0, stays there if held
                    strt, // exposure start, actual may be delayed
                    a,    // [ 7:0] external program memory address
                    re,   // enable program memory read (0 - hold)
                    d,    // [15:0] external program memory data
                    v_ph, // [ 2:0] vertical phase
                    v_stb,// vertical phase changed
                    expos,// exposure in progress
                    frame,// frame readout in progress
                    line, // line readout in progress (active hor. phases)
                    fd,   // fast drain
                    shtr, // electronic shutter
                    shtrmode, // 1 - use mechanical shutter (first pair of toggles)
                    mshutter  // mechanical shutter pulse output
                    );// shutter
    input clk;
    input en;
    input strt;
    output [7:0] a;
    output re;
    input [15:0] d;
    output [2:0] v_ph;
    output v_stb;
    output expos;
    output frame;
    output line;
    output fd;
    output shtr;
    input  shtrmode;
    output mshutter;

/* command table:
New modification for mechanical shutter. First shutter pulse is mechanical shutter pulse,
second (and all the next if present) - electronic shutter

   0000 NNNN NNNN NNNN - begin repeat loop (up to 4095 times). 0000 0000 0000 - forever
   0001 00xx xxxx xxxx - end loop
   0001 01xx aaaa aaaa - jump (aaaa aaaa), pop loop stack (jump out of the loop)
   0001 10xx aaaa aaaa - jump (aaaa aaaa)
   0001 11xx aaaa aaaa - jump (aaaa aaaa) if start pending, pop loop stack
   0010 wwww wwww wwww - reset toggle signals, wait (wwww wwww wwww) cycles
   0011 wwww wwww wwww - toggle "expos", wait
   0100 wwww wwww wwww - toggle "frame", wait
   0101 wwww wwww wwww - toggle "line", wait
   0110 wwww wwww wwww - toggle "fd", wait
   0110 wwww wwww wwww - toggle "shutter", wait
   1000 wwww wwww wwww - set vertical phase #0 V1M V2L
   1001 wwww wwww wwww - set vertical phase #1 V1L V2M
   1010 wwww wwww wwww - set vertical phase #2 V1M V2M
   1011 wwww wwww wwww - set vertical phase #3 V1L V2H
   1100 wwww wwww wwww - set vertical phase #4 V1L V2L
   1101 wwww wwww wwww - set vertical phase #5 V1M V2A
   1110 wwww wwww wwww - set vertical phase #6 V1A V2M
   1111 wwww wwww wwww - set vertical phase #7 ?   ?
*/
  reg rst0;
  wire rst;
  reg stack_err;
   
  wire  cmd_loop=            (d[15:12]==  4'h0);
  wire  cmd_endloop=         (d[15:10]== {4'h1,2'b00});
  wire  cmd_jumpout=         (d[15:10]== {4'h1,2'b01});
  wire  cmd_jump=            (d[15:10]== {4'h1,2'b10});
  wire  cmd_cjump=           (d[15:10]== {4'h1,2'b11});
  wire  cmd_rst=             (d[15:12]==  4'h2);
  wire  cmd_toggle_expos=    (d[15:12]==  4'h3);
  wire  cmd_toggle_frame=    (d[15:12]==  4'h4);
  wire  cmd_toggle_line=     (d[15:12]==  4'h5);
  wire  cmd_toggle_fd=       (d[15:12]==  4'h6);
  wire  cmd_toggle_shtr=     (d[15:12]==  4'h7);
  wire  cmd_wr_v_ph=          d[15];
  wire  cmd_wait=            (d[15] || d[14] || d[13]) && (d[11:0]!=12'b0);
  
  reg [1:0] shtr_cntr; // 0, 1 - mechanical, 2 - off , 3 - electronic -> 2 -> 3 ->2 ->3
  wire      shtr=     (shtr_cntr[1:0]==2'h3);
  wire      mshutter= (shtr_cntr[1:0]==2'h1);
  
  wire [ 7:0] a;
  reg  [ 7:0] cur_addr;
  wire [ 7:0] cur_addr_plus_1=cur_addr[7:0]+1;
  reg  [11:0] wcntr;
  reg         was_cmd_wait;
  wire        re;
  reg         was_re;
  wire [11:0] loop_cntr;
  wire [ 7:0] loop_addr;
  reg         start_pend;
//  wire [ 3:0] stack_addr;
  reg  [ 3:0] cur_stack_addr;
  wire        loop_cntr_zero= (loop_cntr[11:0]==12'b0);
  wire [11:0] loop_cntr_minus_1=loop_cntr[11:0]-1;

  reg [2:0] v_ph;
  reg       v_stb;
  reg       expos;
  reg       frame;
  reg       line;
  reg       fd;
//  reg       shtr;

  always @ (posedge clk) rst0 <= !en;
  always @ (posedge clk)
   if (!en) stack_err <= 1'b0;
   else if (cur_stack_addr[3:0]==4'hf) stack_err <= 1'b1;
  assign rst=rst0 || stack_err;

  assign re= en && (rst || !cmd_wait || (was_cmd_wait && (wcntr==12'h001)));
  assign a= {8{!rst}} &
       (cmd_endloop?
            ((loop_cntr[11:0]==12'h001)?(cur_addr_plus_1):(loop_addr[ 7:0])):
       ((cmd_jump || cmd_jumpout || (cmd_cjump && start_pend))?(d[7:0]):(cur_addr_plus_1))
       );
//  assign stack_addr[7:0]=cur_stack_addr[7:0]+ cmd_loop; //
  always @ (posedge clk) begin
    was_cmd_wait <= cmd_wait;
    was_re       <= re;
    if (((d[15] || d[14] || d[13])) && was_re) wcntr[11:0] <= d[11:0];
    else wcntr[11:0] <= wcntr[11:0]-1;
    if (strt)   start_pend <= !rst;
    else if (re) start_pend <= !rst && !cmd_cjump && start_pend;
    if (re) cur_addr[7:0] <= a;

    if      (!en)                                                                                     cur_stack_addr[3:0] <= 4'b0;
    else if (cmd_loop)                                                                                cur_stack_addr[3:0] <= cur_stack_addr[3:0]+1;
    else if (cmd_jumpout || (cmd_cjump && start_pend) || (cmd_endloop && (loop_cntr[11:0]==12'h001))) cur_stack_addr[3:0] <= cur_stack_addr[3:0]-1;

    if      (rst)               v_ph[2:0] <= 3'b0;
    else if (was_re && cmd_wr_v_ph) v_ph[2:0] <= d[14:12];
    v_stb <= (was_re && cmd_wr_v_ph) || (en && rst);
    if      (rst || (was_re && cmd_rst)) expos <= 1'b0;
    else if (was_re && cmd_toggle_expos) expos <= !expos;
    if      (rst || (was_re && cmd_rst)) frame <= 1'b0;
    else if (was_re && cmd_toggle_frame) frame <= !frame;
    if      (rst || (was_re && cmd_rst)) line  <= 1'b0;
    else if (was_re && cmd_toggle_line ) line  <= !line;
    if      (rst || (was_re && cmd_rst)) fd    <= 1'b0;
    else if (was_re && cmd_toggle_fd   ) fd    <= !fd;
    
//    input  shtrmode;
//    output mshutter;
//  reg [1:0] shtr_cntr; // 0, 1 - mechanical, 2 - off , 3 - electronic -> 2 -> 3 ->2 ->3
    
    if      (rst || (was_re && cmd_rst)) shtr_cntr[1:0]  <= {~shtrmode,1'b0};
    else if (was_re && cmd_toggle_shtr ) shtr_cntr[1:0]  <= {|shtr_cntr[1:0],~shtr_cntr[0]};
    
//    if      (rst || (was_re && cmd_rst)) shtr  <= 1'b0;
//    else if (was_re && cmd_toggle_shtr ) shtr  <= !shtr;

  end


  RAM16XnD # (8) i_loop_addr(.D(cur_addr_plus_1),
                              .WE(cmd_loop),
                              .CLK(clk),
                              .AW(cur_stack_addr[3:0]+ cmd_loop), // when "push" - pre increment
                              .AR(4'b0),
                              .QW(loop_addr[ 7:0]),
                              .QR());
  RAM16XnD #(12) i_loop_cntr(.D(cmd_loop?d[11:0]:loop_cntr_minus_1[11:0]),
                              .WE(cmd_loop || (cmd_endloop && !loop_cntr_zero)), // for "forever" loop
                              .CLK(clk),
                              .AW(cur_stack_addr[3:0]+ cmd_loop),  // when "push" - pre increment
                              .AR(4'b0),
                              .QW(loop_cntr[11:0]),
                              .QR());
endmodule
