/*
** -----------------------------------------------------------------------------**
** shutter_seq.v
**
** Programmable sequencer for the mechanical shutter
**
** Copyright (C) 2007 Elphel, Inc.
**
** -----------------------------------------------------------------------------**
**  This file is part of x347
**  x347 is free software - hardware description language (HDL) code.
** 
**  This program is free software: you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation, either version 3 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program.  If not, see <http://www.gnu.org/licenses/>.
** -----------------------------------------------------------------------------**
**
**  $Log: shutter_seq.v,v $
**  Revision 1.1.1.1  2008/11/27 20:04:02  elphel
**
**
**  Revision 1.1.1.1  2007/10/10 03:51:49  elphel
**  This is a fresh tree based on elphel353-2.10
**
**  Revision 1.6  2007/08/17 12:07:19  spectr_rain
**  switch to GPL3 license
**
**  Revision 1.5  2007/07/20 09:46:13  spectr_rain
**  *** empty log message ***
**
**  Revision 1.2  2007/05/25 02:09:09  elphel
**  347 update for mechanical shutter
**
**
*/

module shutter_seq (clk, // "pixel" clock (25MHz)
                    wr,  // write enable, sync to clk
                    wa,  // wa
                    wd,  // 16 bit data 
                    en,  // enable sequencer (0 immediately resets output to 8'h0)
                    alt,  // alternate odd/even exposures
                    strt, // start sequencer (OK to have it > 1 cycle, will use leading edge)
                    run,  // sequencer is running
                    q);   // 8-bit output (idle - 8'h0)                 


  input         clk;
  input         wr;
  input   [7:0] wa;
  input  [15:0] wd;
  input         en;
  input         alt;
  input         strt;
  output        run;
  output  [7:0] q;

  reg    [7:0] q;
  reg    [7:0] pgm_a;
  wire  [31:0] pgm_d;
  wire  [15:0] ramo;
  reg   [15:0] ramo_r;
  reg          run, wasrun;
  assign pgm_d[31:0] = {ramo[15:0],ramo_r[15:0]};
//  reg   [4:0]  phase;
  reg   [3:0]  phase;
  reg          pgm_re;  // 2-cycle long
  reg          pgm_re2; // second cycle (reading the MSB)
  reg          got_data; // data in pgm_d just got valid
  reg          next_instr;
  reg   [19:0] dly_cntr;
  wire         phase_last;
  reg          strt_d;
  reg          even_exp; // even exposure (starts from 1-st, not zero-th)
  
  wire         start_single;
  assign       phase_last=phase[3]; // change to [3] for /4 cycle
  assign       start_single=strt & ~strt_d;
//     
 RAMB16_S18_S18 i_shutterseq (
      .DOA(),                   // Port A 16-bit Data Output
      .DOPA(),                  // Port A 2-bit Parity Output
      .ADDRA({2'h0,wa[7:0]}),   // Port A 10-bit Address Input
      .CLKA(clk),               // Port A Clock
      .DIA(wd[15:0]),           // Port A 16-bit Data Input
      .DIPA(2'h0),              // Port A 2-bit parity Input
      .ENA(wr),                 // Port A RAM Enable Input
      .SSRA(1'b0),              // Port A Synchronous Set/Reset Input
      .WEA(1'b1),               // Port A Write Enable Input

      .DOB(ramo[15:0]),        // Port B 16-bit Data Output
      .DOPB(),                  // Port B 2-bit Parity Output
      .ADDRB({2'h0,pgm_a[7:0]}),// Port B 10-bit Address Input
      .CLKB(clk),               // Port B Clock
      .DIB(16'h0),              // Port B 16-bit Data Input
      .DIPB(2'h0),              // Port B 2-bit parity Input
      .ENB(pgm_re),             // Port B RAM Enable Input
      .SSRB(1'b0),              // Port B Synchronous Set/Reset Input
      .WEB(1'b0)                // Port B Write Enable Input
   );
/* Simple sequencer, no loops, up to 128 commands.
   Each command is 32-bit word (uses 2 of 16-bit ones from the interface - little endian):
   0XXX XXXX XXXX XXXX XXXX XXXX XXXX XXXX - end sequencer
   1XX0 XXXX XXXX NNNN NNNN NNNN NNNN NNNN - nop, just wait N..N clocks
   1X1X XXXX XXXX NNNN NNNN NNNN NNNN NNNN - nop, just wait N..N clocks
   11XX XXXX XXXX NNNN NNNN NNNN NNNN NNNN - nop, just wait N..N clocks
   1001 BBBB BBBB NNNN NNNN NNNN NNNN NNNN - set output to B..B, wait N..N clocks
   N..N==0 - command takes one cycle
   1 cycle is made of 4  clock cycles (pixel clocks)
*/
  always @ (posedge clk) begin
  
    if      (!alt)         even_exp <= 1'b0;
	 else if (start_single) even_exp <= ~even_exp;
    strt_d <= strt;

    phase[3:0] <= run?{phase[2:0],phase_last}:{3'b0,start_single};
    run <= en && (run? (!(got_data && !pgm_d[31])) : start_single);
    got_data <= pgm_re && phase[1];
    next_instr <= !run || (phase[2]? (got_data?(pgm_d[19:0]==20'h0):(dly_cntr[19:1]==19'h0)):next_instr);
    pgm_re2 <= pgm_re && !pgm_re2;
    pgm_re <= en && (pgm_re? (!pgm_re2):( run? (next_instr && phase_last ): start_single));
    pgm_a[7:0] <= run ? (pgm_a[7:0] + pgm_re) : {even_exp,7'h0};
    if (phase[2]) dly_cntr[19:0] <= got_data? pgm_d[19:0]:(dly_cntr[19:0]-1);
    if      (!run)                             q[7:0] <= 8'h0;
    else if (got_data && (pgm_d[30:28]==3'h1)) q[7:0] <= pgm_d[27:20];
    
    ramo_r[15:0] <= ramo[15:0];
  end

endmodule
